VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cSqlite"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False

'Autor          : J. Elihu
'----------------------------------------------------
'Version        : 1.6
'Date           : 14/12/20

'Requirements   :
'               - cLib
'               - cSqlite
'               - cSqlitecursor

'Credits        :
'                - Cobein   (mSqlite)
'                - Cocus    (cSQLiteConnection)

Option Explicit

Private Declare Function WideCharToMultiByte Lib "kernel32" (ByVal CodePage As Long, ByVal dwFlags As Long, ByVal lpWideCharStr As Long, ByVal cchWideChar As Long, ByRef lpMultiByteStr As Any, ByVal cchMultiByte As Long, ByVal lpDefaultChar As String, ByVal lpUsedDefaultChar As Long) As Long
Private Declare Function MultiByteToWideChar Lib "kernel32" (ByVal CodePage As Long, ByVal dwFlags As Long, ByVal lpMultiByteStr As Long, ByVal cchMultiByte As Long, ByVal lpWideCharStr As Long, ByVal cchWideChar As Long) As Long
Private Declare Function lstrlen Lib "kernel32" Alias "lstrlenW" (ByVal lpString As Long) As Long
Private Declare Function lstrlenA Lib "kernel32" (ByVal lpString As Long) As Long
Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (pDst As Any, pSrc As Any, ByVal ByteLen As Long)

Private Type tDataBlob
    lPtr    As Long
    lSize   As Long
End Type

Public Enum Sqlite3Result
    SQLITE_OK = 0           ' Successful result
    SQLITE_ERROR = 1        ' SQL error or missing database
    SQLITE_INTERNAL = 2     ' Internal logic error in SQLite
    SQLITE_PERM = 3         ' Access permission denied
    SQLITE_ABORT = 4        ' Callback routine requested an abort
    SQLITE_BUSY = 5         ' The database file is locked
    SQLITE_LOCKED = 6       ' A table in the database is locked
    SQLITE_NOMEM = 7        ' A malloc() failed
    SQLITE_READONLY = 8     ' Attempt to write a readonly database
    SQLITE_INTERRUPT = 9    ' Operation terminated by sqlite3_interrupt()
    SQLITE_IOERR = 10       ' Some kind of disk I/O error occurred
    SQLITE_CORRUPT = 11     ' The database disk image is malformed
    SQLITE_NOTFOUND = 12    ' NOT USED. Table or record not found
    SQLITE_FULL = 13        ' Insertion failed because database is full
    SQLITE_CANTOPEN = 14    ' Unable to open the database file
    SQLITE_PROTOCOL = 15    ' NOT USED. Database lock protocol error
    SQLITE_EMPTY = 16       ' Database is empty
    SQLITE_SCHEMA = 17      ' The database schema changed
    SQLITE_TOOBIG = 18      ' String or BLOB exceeds size limit
    SQLITE_CONSTRAINT = 19  ' Abort due to constraint violation
    SQLITE_MISMATCH = 20    ' Data type mismatch
    SQLITE_MISUSE = 21      ' Library used incorrectly
    SQLITE_NOLFS = 22       ' Uses OS features not supported on host
    SQLITE_AUTH = 23        ' Authorization denied
    SQLITE_FORMAT = 24      ' Auxiliary database format error
    SQLITE_RANGE = 25       ' 2nd parameter to sqlite3_bind out of range
    SQLITE_NOTADB = 26      ' File opened that is not a database file
    SQLITE_ROW = 100        ' sqlite3_step() has another row ready
    SQLITE_DONE = 101       ' sqlite3_step() has finished executing
End Enum

'//Fundamental Datatypes
'Public Enum Sqlite3Datatypes
'    SQLITE_INTEGER = 1
'    SQLITE_FLOAT = 2
'    SQLITE_TEXT = 3
'    SQLITE_BLOB = 4
'    SQLITE_NULL = 5
'End Enum

'// Authorizer Return Codes
'Public Enum Sqlite3AuthorizerCodes
'    SQLITE_DENY = 1     ' Abort the SQL statement with an error */
'    SQLITE_IGNORE = 2   ' Don't allow access, but don't generate an error */
'End Enum

Public Enum enmSqlitePlugin
    [Sql3_Library]
    [Sql3_Startup]
    [Sql3_Shutdown]
End Enum
Public Enum enmSqliteEncoding
    [UTF_8]
    [UTF_16]
    [UTF_16le]
    [UTF_16be]
End Enum
Public Enum enmSqliteSynchronous
    [OFF] = 0
    [NORMAL_] = 1
    [FULL] = 2
    [EXTRA] = 3
End Enum
Public Enum enmSqliteJournalMode
   [DELETE_] = 0
   [TRUNCATE] = 1
   [PERSIST] = 2
   [MEMORY] = 3
   [WAL] = 4
   [OFF] = 5
End Enum

Private pSqlite3    As Long


Private Sub Class_Initialize()
    Sqlite3Manage 1
End Sub
Private Sub Class_Terminate()
    If pSqlite3 Then Call SqliteClose
    Sqlite3Manage 2
End Sub

Public Function SqliteOpen(ByVal FileName As String) As Sqlite3Result
Dim bvStr() As Byte
    If pSqlite3 Then Call SqliteClose
    bvStr = ppString(FileName)
    SqliteOpen = oSql3_dll.Func("sqlite3_open", VarPtr(bvStr(0)), VarPtr(pSqlite3))
End Function
Public Function SqliteClose() As Sqlite3Result
    SqliteClose = oSql3_dll.Func("sqlite3_close", pSqlite3)
    If SqliteClose = SQLITE_OK Then pSqlite3 = 0
End Function
Public Function Execute(ByVal Sql As String) As Sqlite3Result
    Execute = sqlite3_exec(Sql)
End Function
Public Function Query(ByVal Sql As String, Optional Result As Sqlite3Result) As cSqliteCursor
Dim eStmt  As Long
    Set Query = New cSqliteCursor
    Result = sqlite3_prepare_v2(Sql, eStmt)
    If Result = SQLITE_OK Then Call Query.Link(eStmt, pSqlite3)
End Function

Public Function Transaction(ByRef Queries As Collection) As Sqlite3Result
Dim pSql    As String
Dim lResult As Long
Dim i       As Long

    If Not (pSqlite3 <> 0) Then Exit Function
    lResult = sqlite3_exec("BEGIN;")
    If lResult <> SQLITE_OK Then GoTo e
    For i = 1 To Queries.Count
        pSql = Queries.Item(i)
        lResult = sqlite3_exec(pSql)
        If lResult <> SQLITE_OK Then GoTo e
    Next
    lResult = sqlite3_exec("END;")
e:
 Transaction = lResult
End Function

Public Function Attach(ByVal FileName As String, ByVal DataBaseName As String) As Sqlite3Result
   Attach = Execute("ATTACH DATABASE '" & FileName & "' AS '" & DataBaseName & "';")
End Function
Public Function Detach(ByVal DataBaseName As String) As Sqlite3Result
   Detach = Execute("DETACH DATABASE '" & DataBaseName & "';")
End Function

Public Function SqliteErrmsg() As String
    If oSql3_dll.Handle = 0 Then: SqliteErrmsg = "Not loaded Library ": Exit Function
    SqliteErrmsg = StringFromPtr(oSql3_dll.Func("sqlite3_errmsg", pSqlite3))
End Function

Public Function SqliteErrcode() As Sqlite3Result
    SqliteErrcode = oSql3_dll.Func("sqlite3_errcode", pSqlite3)
End Function
Public Function LastInsertID() As Long
    LastInsertID = oSql3_dll.Func("sqlite3_last_insert_rowid", pSqlite3)
End Function

Property Get Handle() As Long: Handle = pSqlite3: End Property
Property Get SqliteChanges() As Long
    SqliteChanges = oSql3_dll.Func("sqlite3_changes", pSqlite3)
End Property
Property Get SqliteVersion() As String
    SqliteVersion = StringFromPtr(oSql3_dll.Func("sqlite3_libversion"))
End Property
Property Get SqliteVersionNum() As Long
    SqliteVersionNum = oSql3_dll.Func("sqlite3_libversion_number")
End Property
Property Get File() As String
    File = StringFromPtr(oSql3_dll.Func("sqlite3_db_filename", pSqlite3, 0&))
End Property

Property Get JournalMode() As enmSqliteJournalMode
Dim tmp  As String
Dim Stmt As Long

    If pSqlite3 = 0 Then Exit Property
    If sqlite3_prepare_v2("PRAGMA journal_mode;", Stmt) Then Exit Property
    If sqlite3_step(Stmt) = SQLITE_ROW Then
        tmp = sqlite3_column_text16(Stmt, 0)
        Select Case LCase(tmp)
            Case "delete": JournalMode = 0
            Case "truncate": JournalMode = 1
            Case "persist": JournalMode = 2
            Case "memory": JournalMode = 3
            Case "wal": JournalMode = 4
            Case "off": JournalMode = 5
        End Select
    End If
    Call sqlite3_finalize(Stmt)
End Property
Property Let JournalMode(ByVal Value As enmSqliteJournalMode)
    If pSqlite3 = 0 Then Exit Property
    Select Case Value
        Case 0: sqlite3_exec "PRAGMA journal_mode = DELETE;"
        Case 1: sqlite3_exec "PRAGMA journal_mode = TRUNCATE;"
        Case 2: sqlite3_exec "PRAGMA journal_mode = PERSIST;"
        Case 3: sqlite3_exec "PRAGMA journal_mode = MEMORY;"
        Case 4: sqlite3_exec "PRAGMA journal_mode = WAL;"
        Case 5: sqlite3_exec "PRAGMA journal_mode = OFF;"
    End Select
End Property

Property Get JournalSizeLimit() As Long
Dim Stmt As Long
    If pSqlite3 = 0 Then Exit Property
    If sqlite3_prepare_v2("PRAGMA journal_size_limit;", Stmt) Then Exit Property
    If sqlite3_step(Stmt) = SQLITE_ROW Then
        JournalSizeLimit = sqlite3_column_int(Stmt, 0)
    End If
    Call sqlite3_finalize(Stmt)
End Property
Property Let JournalSizeLimit(ByVal Value As Long)
    If pSqlite3 = 0 Then Exit Property
    sqlite3_exec "PRAGMA journal_size_limit = " & Value & ";"
End Property

Property Get Synchronous() As enmSqliteSynchronous
Dim Stmt As Long
    If pSqlite3 = 0 Then Exit Property
    If sqlite3_prepare_v2("PRAGMA synchronous;", Stmt) Then Exit Property
    If sqlite3_step(Stmt) = SQLITE_ROW Then
        Synchronous = sqlite3_column_int(Stmt, 0)
    End If
    Call sqlite3_finalize(Stmt)
End Property
Property Let Synchronous(ByVal Value As enmSqliteSynchronous)
    If pSqlite3 = 0 Then Exit Property
    sqlite3_exec "PRAGMA synchronous = " & Value & ";"
End Property

Property Get ForeignKeys() As Boolean
Dim Stmt As Long
    If pSqlite3 = 0 Then Exit Property
    If sqlite3_prepare_v2("PRAGMA foreign_keys;", Stmt) Then Exit Property
    If sqlite3_step(Stmt) = SQLITE_ROW Then
        ForeignKeys = sqlite3_column_int(Stmt, 0) <> 0
    End If
    Call sqlite3_finalize(Stmt)
End Property
Property Let ForeignKeys(ByVal Value As Boolean)
    If pSqlite3 = 0 Then Exit Property
    sqlite3_exec "PRAGMA foreign_keys  = " & IIf(Value, 1, 0) & ";"
End Property

Property Get Encoding() As enmSqliteEncoding
Dim Stmt As Long
    If pSqlite3 = 0 Then Exit Property
    If sqlite3_prepare_v2("PRAGMA encoding;", Stmt) Then Exit Property
    If sqlite3_step(Stmt) = SQLITE_ROW Then
        Select Case LCase(sqlite3_column_text16(Stmt, 0))
            Case "utf-8": Encoding = 0
            Case "utf-16": Encoding = 1
            Case "utf-16le": Encoding = 2
            Case "utf-16be": Encoding = 3
        End Select
    End If
    Call sqlite3_finalize(Stmt)
End Property
Property Let Encoding(ByVal Value As enmSqliteEncoding)
    If pSqlite3 = 0 Then Exit Property
    Select Case Value
        Case 0: sqlite3_exec "PRAGMA encoding = UTF-8;"
        Case 1: sqlite3_exec "PRAGMA encoding = UTF-16;"
        Case 2: sqlite3_exec "PRAGMA encoding = UTF-16le;"
        Case 3: sqlite3_exec "PRAGMA encoding = UTF-16be;"
    End Select
End Property
Property Let UserVersion(ByVal Value As Integer)
    If pSqlite3 = 0 Then Exit Property
    sqlite3_exec "PRAGMA user_version = " & Value & ";"
End Property
Property Get UserVersion() As Integer
Dim Stmt As Long
    If pSqlite3 = 0 Then Exit Property
    If sqlite3_prepare_v2("PRAGMA user_version;", Stmt) Then Exit Property
    If sqlite3_step(Stmt) = SQLITE_ROW Then
        UserVersion = sqlite3_column_text16(Stmt, 0)
    End If
    Call sqlite3_finalize(Stmt)
End Property
Public Function Vacuum() As Sqlite3Result
    Vacuum = sqlite3_exec("VACUUM;")
End Function


'\** Sqlite Backup CallBack **/ -> Include in [Class || Form]
'Public Sub Sqlite_Backup(Remaining As Long, PageCount As Long, ByVal Result As Sqlite3Result, ByVal zMsg As String)
'   Debug.Print "Backup: " & 100 * (PageCount - Remaining) \ PageCount & "%"
'End Sub

Public Function Backup(ByVal FileNameBackup As String, Optional ObjCallBack As Object, Optional PageStep As Long = -1, Optional CallDoEvents As Boolean, Optional ByVal dbSourceName As String = "main", Optional dbDestName As String = "main") As Sqlite3Result
On Local Error Resume Next
Dim bvStr1()  As Byte
Dim bvStr2()  As Byte
Dim pBackupDb As Long
Dim pBackup3  As Long
Dim lResult   As Sqlite3Result


    If pSqlite3 = 0 Then Exit Function
    
    bvStr1 = ppString(FileNameBackup)
    lResult = oSql3_dll.Func("sqlite3_open", VarPtr(bvStr1(0)), VarPtr(pBackupDb))
    If lResult <> SQLITE_OK Then
        If Not ObjCallBack Is Nothing Then Call ObjCallBack.Sqlite_Backup(0, 0, lResult, StringFromPtr(oSql3_dll.Func("sqlite3_errmsg", pBackupDb)))
        GoTo e:
    End If

    bvStr1 = ppString(dbSourceName)
    bvStr2 = ppString(dbDestName)
    
    pBackup3 = oSql3_dll.Func("sqlite3_backup_init", pBackupDb, VarPtr(bvStr2(0)), pSqlite3, VarPtr(bvStr1(0)))
    
    If (pBackup3) Then
        
        Do
        
            lResult = oSql3_dll.Func("sqlite3_backup_step", pBackup3, PageStep)
            If Not ObjCallBack Is Nothing And (lResult = SQLITE_OK Or lResult = SQLITE_DONE) Then
                Call ObjCallBack.Sqlite_Backup(oSql3_dll.Func("sqlite3_backup_remaining", pBackup3), oSql3_dll.Func("sqlite3_backup_pagecount", pBackup3), lResult, "")
            End If
            
            If (lResult = SQLITE_OK Or lResult = SQLITE_BUSY Or lResult = SQLITE_LOCKED) Then Call oSql3_dll.Func("sqlite3_sleep", 250)
            If CallDoEvents Then DoEvents
            
        Loop While (lResult = SQLITE_OK Or lResult = SQLITE_BUSY Or lResult = SQLITE_LOCKED)
        
        lResult = oSql3_dll.Func("sqlite3_backup_finish", pBackup3)
        If lResult <> SQLITE_OK Then
            If Not ObjCallBack Is Nothing Then Call ObjCallBack.Sqlite_Backup(0, 0, lResult, StringFromPtr(oSql3_dll.Func("sqlite3_errmsg", pBackupDb)))
        End If
        
    Else
        lResult = oSql3_dll.Func("sqlite3_errcode", pBackupDb)
        If Not ObjCallBack Is Nothing Then Call ObjCallBack.Sqlite_Backup(0, 0, lResult, StringFromPtr(oSql3_dll.Func("sqlite3_errmsg", pBackupDb)))
    End If
    
e:
    If pBackupDb Then Call oSql3_dll.Func("sqlite3_close", pBackupDb)
    Backup = lResult
End Function





'---------------------------------------------------------------------------
'                          Private Subs
'---------------------------------------------------------------------------

Private Function ppString(Data As String) As Byte()
    ppString = StrConv(UnicodeToUTF8(Data) & vbNullChar, vbFromUnicode)
End Function
Private Function sqlite3_prepare_v2(ByVal zSQL As String, ByRef ppStmt As Long) As Long
Dim bvStr() As Byte
    bvStr = ppString(zSQL)
    sqlite3_prepare_v2 = oSql3_dll.Func("sqlite3_prepare_v2", pSqlite3, VarPtr(bvStr(0)), Len(zSQL), VarPtr(ppStmt), 0&)
End Function
Private Function sqlite3_step(ByVal Stmt As Long) As Sqlite3Result
    sqlite3_step = oSql3_dll.Func("sqlite3_step", Stmt)
End Function
Private Function sqlite3_column_text(ByVal Stmt As Long, ByVal Col As Long) As String
Dim eBlob As tDataBlob
    eBlob.lPtr = oSql3_dll.Func("sqlite3_column_text", Stmt, Col)
    eBlob.lSize = sqlite3_column_bytes(Stmt, Col)
    sqlite3_column_text = BytesFromPtr(eBlob.lPtr, eBlob.lSize)
End Function
Private Function sqlite3_column_text16(ByVal Stmt As Long, ByVal Col As Long) As String
Dim eBlob As tDataBlob
    eBlob.lPtr = oSql3_dll.Func("sqlite3_column_text16", Stmt, Col)
    eBlob.lSize = sqlite3_column_bytes16(Stmt, Col)
    sqlite3_column_text16 = BytesFromPtr(eBlob.lPtr, eBlob.lSize)
End Function
Private Function sqlite3_column_int(ByVal Stmt As Long, ByVal Col As Long) As Long
        sqlite3_column_int = oSql3_dll.Func("sqlite3_column_int", Stmt, Col)
End Function
Private Function sqlite3_finalize(ByVal Stmt As Long) As Sqlite3Result
    sqlite3_finalize = oSql3_dll.Func("sqlite3_finalize", Stmt)
End Function
Private Function sqlite3_column_bytes(ByVal Stmt As Long, ByVal Col As Long) As Long
    sqlite3_column_bytes = oSql3_dll.Func("sqlite3_column_bytes", Stmt, Col)
End Function
Private Function sqlite3_column_bytes16(ByVal Stmt As Long, ByVal Col As Long) As Long
    sqlite3_column_bytes16 = oSql3_dll.Func("sqlite3_column_bytes16", Stmt, Col)
End Function
Private Function sqlite3_exec(ByVal zSQL As String) As Long
Dim bvStr() As Byte
    bvStr = ppString(zSQL)
    sqlite3_exec = oSql3_dll.Func("sqlite3_exec", pSqlite3, VarPtr(bvStr(0)), 0, 0, 0)
End Function


Private Function test_pragma()
Dim eSmt As Long

    If pSqlite3 = 0 Then Exit Function
    If sqlite3_prepare_v2("PRAGMA query_only;", eSmt) Then Exit Function
    If sqlite3_step(eSmt) = SQLITE_ROW Then
        Debug.Print sqlite3_column_text16(eSmt, 0)
    End If
    Call sqlite3_finalize(eSmt)
    
End Function

Private Function StringFromPtr(Ptr As Long, Optional ByVal Size As Long = -1) As String
Dim pzLen   As Long

    If Ptr = 0 Then Exit Function
    If Size = -1 Then Size = lstrlenA(Ptr)
    If Not Size > 0 Then Exit Function
    pzLen = MultiByteToWideChar(65001, 0, Ptr, Size, 0, 0)
    If Not pzLen > 0 Then Exit Function

    StringFromPtr = Space$(pzLen)
    MultiByteToWideChar 65001, 0, Ptr, Size, StrPtr(StringFromPtr), pzLen
    
End Function

Private Function BytesFromPtr(ByVal lAddr As Long, ByVal lSize As Long) As Byte()
On Error GoTo Err
Dim bvData() As Byte
    ReDim bvData(lSize - 1)
    CopyMemory bvData(0), ByVal lAddr, lSize - 1
    BytesFromPtr = bvData
Err:
End Function

Private Function UnicodeToUTF8(ByVal sData As String) As String
Dim bvData()    As Byte
Dim lSize       As Long
Dim lRet        As Long
    If LenB(sData) Then
        lSize = Len(sData) * 4
        ReDim bvData(lSize)
        lRet = WideCharToMultiByte(65001, 0, StrPtr(sData), Len(sData), bvData(0), lSize + 1, vbNullString, 0)
        If lRet Then
            ReDim Preserve bvData(lRet - 1)
            UnicodeToUTF8 = StrConv(bvData, vbUnicode)
        End If
    End If
End Function


